/*
  Copyright (C) 2016-2026 Dan Casarin (https://www.kfrlib.com)
  This file is part of KFR

  KFR is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.

  KFR is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with KFR.

  If GPL is not suitable for your project, you must purchase a commercial license to use KFR.
  Buying a commercial license is mandatory as soon as you develop commercial activities without
  disclosing the source code of your own applications.
  See https://www.kfrlib.com for details.
 */
#pragma once

#include "math.hpp"

#include "base/basic_expressions.hpp"
#include "base/conversion.hpp"
#include "base/endianness.hpp"
#include "base/expression.hpp"
#include "base/filter.hpp"
#include "base/fraction.hpp"
#include "base/generators.hpp"
#include "base/handle.hpp"
#include "base/inline_vector.hpp"
#include "base/math_expressions.hpp"
#include "base/memory.hpp"
#include "base/npy.hpp"
#include "base/random.hpp"
#include "base/random_bits.hpp"
#include "base/reduce.hpp"
#include "base/shape.hpp"
#include "base/simd_expressions.hpp"
#include "base/small_buffer.hpp"
#include "base/state_holder.hpp"
#include "base/tensor.hpp"
#include "base/transpose.hpp"
#include "base/univector.hpp"
