/*
  Copyright (C) 2016-2026 Dan Casarin (https://www.kfrlib.com)
  This file is part of KFR

  KFR is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.

  KFR is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with KFR.

  If GPL is not suitable for your project, you must purchase a commercial license to use KFR.
  Buying a commercial license is mandatory as soon as you develop commercial activities without
  disclosing the source code of your own applications.
  See https://www.kfrlib.com for details.
 */
#pragma once

#include "simd.hpp"

#include "math/asin_acos.hpp"
#include "math/atan.hpp"
#include "math/compiletime.hpp"
#include "math/complex_math.hpp"
#include "math/gamma.hpp"
#include "math/hyperbolic.hpp"
#include "math/interpolation.hpp"
#include "math/log_exp.hpp"
#include "math/modzerobessel.hpp"
#include "math/sin_cos.hpp"
#include "math/sqrt.hpp"
#include "math/tan.hpp"
